﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.OleDb;
using System.Drawing;
using System.Text;
using System.Globalization;
using System.Windows.Forms;

namespace QueryDesigner
{
    public partial class ParametersDialog : Form
    {
        //------------------------------------------------------------------------------
        #region ** fields

        CultureInfo _invariant = CultureInfo.InvariantCulture;
        DateTimeStyles _dateStyle = DateTimeStyles.AllowInnerWhite | DateTimeStyles.AllowLeadingWhite | DateTimeStyles.AllowTrailingWhite | DateTimeStyles.AllowWhiteSpaces | DateTimeStyles.AssumeLocal;
        NumberStyles _numberStyle = NumberStyles.Any;
        Control _focus;

        #endregion

        //------------------------------------------------------------------------------
        #region ** ctors

        public ParametersDialog()
        {
            InitializeComponent();
        }
        public ParametersDialog(List<OleDbParameter> parms) : this()
        {
            var height = _panel.Height;

            foreach (var p in parms)
            {
                // create label
                var lbl = new Label();
                lbl.Text = CleanupName(p.ParameterName);
                lbl.AutoSize = false;
                lbl.Dock = DockStyle.Fill;
                lbl.TextAlign = ContentAlignment.MiddleRight;
                _panel.Controls.Add(lbl);
                _panel.SetRow(lbl, _panel.RowCount - 1);

                // create input control
                Control ctl = GetControl(p);
                ctl.Tag = p;
                ctl.AutoSize = false;
                ctl.Dock = DockStyle.Fill;
                _panel.Controls.Add(ctl);
                _panel.SetRow(ctl, _panel.RowCount - 1);
                _panel.SetColumn(ctl, 1);
                if (_focus == null)
                {
                    _focus = ctl;
                }

                // add another row to the panel
                _panel.RowCount++;
                var style = new RowStyle(SizeType.AutoSize);
                _panel.RowStyles.Add(style);
            }

            // adjust form height
            Height += _panel.Height - height;
        }

        #endregion

        //------------------------------------------------------------------------------
        #region ** implementation

        // set focus to first control on the form
        protected override void OnShown(EventArgs e)
        {
            if (_focus != null)
            {
                _focus.Focus();
                _focus = null;
            }
            base.OnShown(e);
        }

        // creates a control of appropriate type for the parameter.
        // note: parameter values are stored as invariant strings.
        Control GetControl(OleDbParameter p)
        {
            var value = p.Value as string;
            var type = OleDbSchema.GetType(p.OleDbType);
            if (OleDbSchema.IsNumeric(type))
            {
                var num = new NumericUpDown();
                num.Minimum = decimal.MinValue;
                num.Maximum = decimal.MaxValue;
                num.ThousandsSeparator = true;
                decimal dec = 0;
                if (!string.IsNullOrEmpty(value))
                {
                    decimal.TryParse(value, _numberStyle, _invariant, out dec);
                }
                num.Value = dec;
                return num;
            }
            if (type == typeof(DateTime))
            {
                var dtp = new DateTimePicker();
                dtp.Format = p.OleDbType == OleDbType.Filetime
                    ? DateTimePickerFormat.Time
                    : DateTimePickerFormat.Short;
                DateTime dt = DateTime.Now;
                if (!string.IsNullOrEmpty(value))
                {
                    DateTime.TryParse(value, _invariant, _dateStyle, out dt);
                }
                dtp.Value = dt;
                return dtp;
            }
            if (type == typeof(bool))
            {
                var chk = new CheckBox();
                bool b = false;
                if (!string.IsNullOrEmpty(value))
                {
                    bool.TryParse(value as string, out b);
                }
                chk.Checked = b;
                return chk;
            }

            // default: textbox
            var tb = new TextBox();
            tb.Text = value;
            return tb;
        }

        // clean up a name (remove brackets etc)
        static string CleanupName(string name)
        {
            name = name.Replace('_', ' ');
            name = name.Trim();
            var pos = name.LastIndexOf('!');
            if (pos > -1)
            {
                name = name.Substring(pos + 1);
            }
            if (name.Length > 1 && name[0] == '[' && name[name.Length - 1] == ']')
            {
                name = name.Substring(1, name.Length - 2);
            }
            return name;
        }

        // copy values in fields back to parameters
        void _btnOK_Click(object sender, EventArgs e)
        {
            foreach (Control ctl in _panel.Controls)
            {
                var p = ctl.Tag as OleDbParameter;
                if (p != null)
                {
                    CheckBox chk = ctl as CheckBox;
                    if (chk != null)
                    {
                        p.Value = chk.Checked.ToString();
                        continue;
                    }
                    DateTimePicker dtp = ctl as DateTimePicker;
                    if (dtp != null)
                    {
                        p.Value = dtp.Value.ToString(_invariant);
                        continue;
                    }
                    p.Value = ctl.Text;
                }
            }
        }

        #endregion
    }
}
